/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "taucs.h"

typedef struct {
  int     length;
  int*    indices;
  int*    bitmap;
  double* values;
} spa;

#define max(x,y) ( ((x) > (y)) ? (x) : (y) )

/*********************************************************/
/*                                                       */
/*********************************************************/

static spa* spa_create(int n)
{
  int i;
  spa* s;
  
  s = (spa*) malloc( sizeof(spa) );
  if ( !s ) return NULL;

  s->indices = (int*)    malloc( n * sizeof(int) );
  s->bitmap  = (int*)    malloc( n * sizeof(int) );
  s->values  = (double*) malloc( n * sizeof(double) );

  if ( !(s->indices) || !(s->values) || !(s->bitmap) ) {
    taucs_printf("chol: cannot create spa\n");
    free( s->indices );
    free( s->bitmap  );
    free( s->values  );
    free( s );
    return NULL;
  }

  s->length = 0;

  for (i=0; i<n; i++) (s->bitmap)[i] = -1;
  
  return s;
}

static void spa_free(spa* s)
{
  free( s->indices );
  free( s->values  );
  free( s );
}

static void spa_set(spa* s, taucs_ccs_matrix* A, int j)
{
  int i, ip, next;
  double Aij;
  
  assert(j < A->n);

  next = 0;
  for (ip = (A->colptr)[j]; ip < (A->colptr)[j+1]; ip++) {
    i   = (A->rowind)[ip];
    Aij = (A->values)[ip];

    assert( i >= j ); /* A must be lower */
    
    (s->indices)[ next ] = i;
    (s->values) [ i    ] = Aij;
    (s->bitmap) [ i    ] = j;
    next++;
  }

  s->length = next;
}

static void spa_scale_add(spa* s, int j, taucs_ccs_matrix* A, int k, double alpha)
{
  int i, ip, next;
  double Aik;
  
  assert(k < A->n);

  /*
  printf("spa_scale_add: updating column %d with column %d\n",j,k);
  printf("spa_scale_add: colptr %d to %d-1\n",(A->colptr)[k],(A->colptr)[k+1]);
  */

  next = 0;
  for (ip = (A->colptr)[k]; ip < (A->colptr)[k+1]; ip++) {
    i   = (A->rowind)[ip];
    if (i < j) continue;
    Aik = (A->values)[ip];
    
    if ( (s->bitmap)[ i ] < j ) {
      /*printf("fill in (%d,%d)\n",i,j);*/
      (s->bitmap)[i] = j;
      (s->values)[i] = 0.0;
      (s->indices)[ s->length ] = i;
      (s->length)++;
    }

    (s->values)[ i ] += alpha * Aik;

    /*printf("spa_scale_add: A(%d,%d) -= %lg * %lg ==> %lg\n",i,j,alpha,Aik,(s->values)[i]);*/
  }
}
		    
/*********************************************************/
/*                                                       */
/*********************************************************/

int*    rowlist;
int*    rowlist_next;
int*    rowlist_colind;
double* rowlist_values;

int     rowlist_freelist;
int     rowlist_size;

static int rowlist_create(int n)
{
  int i;

  rowlist = (int*) malloc( n * sizeof(int) );

  rowlist_size    = 1000;
  rowlist_next    = (int*)    malloc( rowlist_size * sizeof(int) );
  rowlist_colind = (int*)    malloc( rowlist_size * sizeof(int) );
  rowlist_values  = (double*) malloc( rowlist_size * sizeof(double) );

  for (i=0; i<n; i++) rowlist[i] = -1; /* no list yet for row i */

  /* free list */
  rowlist_freelist = 0; 
  for (i=0; i<rowlist_size-1; i++) rowlist_next[i] = i+1; 
  rowlist_next[rowlist_size-1] = -1;
				   
  return 0;
}

static void rowlist_free()
{
  free(rowlist);
  free(rowlist_next);
  free(rowlist_colind);
  free(rowlist_values);
}

static void rowlist_freerow(int i)
{
}

static void rowlist_add(int i,int j,double v)
{
  int l;

  if (rowlist_freelist == -1) {
    int inc = 1000;
    int ii;

    rowlist_next   = (int*)    realloc( rowlist_next,   (rowlist_size+inc) * sizeof(int) );
    rowlist_colind = (int*)    realloc( rowlist_colind, (rowlist_size+inc) * sizeof(int) );
    rowlist_values = (double*) realloc( rowlist_values, (rowlist_size+inc) * sizeof(double) );

    rowlist_freelist = rowlist_size;
    for (ii=rowlist_size; ii<rowlist_size+inc-1; ii++)
      rowlist_next[ii] = ii+1;
    rowlist_next[ rowlist_size+inc-1 ] = -1;

    rowlist_size    += inc;
  }

  l = rowlist_freelist;
  rowlist_freelist = rowlist_next[ rowlist_freelist ];

  rowlist_next  [ l ] = rowlist[ i ];
  rowlist_colind[ l ] = j;
  rowlist_values[ l ] = v;
  
  rowlist[ i ] = l;
}

static int rowlist_getfirst(int i)
{
  return rowlist[ i ];
}

static int rowlist_getnext(int l)
{
  return rowlist_next[ l ];
}

static int rowlist_getcolind(int l)
{
  return rowlist_colind[ l ];
}

static double rowlist_getvalue(int l)
{
  return rowlist_values[ l ];
}

/*********************************************************/
/* Cholesky factorization                                */
/* This is a left-looking column-column code using       */
/* row lists. Can perform drop-tolerance incomplete      */
/* factorization with or without diagonal modification   */
/* to maintain rowsums.                                  */
/*********************************************************/

/*
symccs_matrix* 
taucs_symccs_factor_llt(symccs_matrix* A,double droptol, int modified)
{return NULL;}
*/

taucs_ccs_matrix* 
taucs_ccs_factor_llt(taucs_ccs_matrix* A,double droptol, int modified)
{
  int            i,j,k,l,n,ip,next,Lnnz;
  double Lkj,pivot,v,norm;
  spa*           s;
  taucs_ccs_matrix* L;
  double* dropped;
  int Aj_nnz;
  double flops = 0.0;

  if (!(A->flags & TAUCS_SYMMETRIC)) {
    taucs_printf("taucs_ccs_factor_llt: matrix must be symmetric\n");
    return NULL;
  }
  if (!(A->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_factor_llt: lower part must be represented\n");
    return NULL;
  }

  n = A->n;

  taucs_printf("taucs_ccs_factor_llt: starting n=%d droptol=%lf modified?=%d\n",
	     n,droptol,modified);

  L = taucs_ccs_create(n,n,1000);
  L->flags = TAUCS_TRIANGULAR | TAUCS_LOWER;

  Lnnz = 1000;
  next = 0;

  s = spa_create(n);
  rowlist_create(n);

  dropped = (double*) malloc( n * sizeof(double) );
  for (i=0; i<n; i++) dropped[i] = 0;

  for (j=0; j<n; j++) {
    spa_set(s,A,j);

    for (l = rowlist_getfirst(j); 
	 l != -1; 
	 l = rowlist_getnext(l)) {
      k   = rowlist_getcolind(l);
      Lkj = rowlist_getvalue(l);
      spa_scale_add(s,j,L,k,-Lkj); /* L_*j -= L_kj * L_*k */
    }

    /* we now add the j'th column of L to the taucs_ccs */
    
    if ( next+(s->length) > Lnnz ) {
      int*    rowind;
      double* values;
      int inc = max( 8192, s->length );
      
      Lnnz += inc;

      rowind = (int*)    realloc( L->rowind, Lnnz * sizeof(int) );
      values = (double*) realloc( L->values, Lnnz * sizeof(double) );
      /* check for errors */
      assert( rowind && values );
      L->rowind = rowind;
      L->values = values;
    }

    (L->colptr)[j] = next;

    norm = 0.0;
    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];
      norm += v*v;
    }
    norm = sqrt(norm);

    Aj_nnz = (A->colptr)[j+1] - (A->colptr)[j];

    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];
      
      if (i==j || fabs(v) > droptol * norm || ip < Aj_nnz) {
	/* if (i==j || fabs(v) > droptol * norm) { */
      } else {
	dropped[i] -= v;
	dropped[j] -= v;
      }
    }


    if (modified) 
      pivot = sqrt( (s->values)[j] - dropped[j] );
    else
      pivot = sqrt( (s->values)[j] );

    if (pivot == 0.0) {
      taucs_printf("taucs_ccs_factor_llt: zero pivot in column %d\n",j);
      taucs_printf("taucs_ccs_factor_llt: Ajj in spa = %lg dropped[j] = %lg Aj_nnz=%d\n",
		 (s->values)[j],dropped[j],Aj_nnz);
    } else if (fabs(pivot) < 1e-12) {
      taucs_printf("taucs_ccs_factor_llt: small pivot in column %d (%le)\n",j,pivot);
    }

    /* we want Lii to be first in the compressed column */
    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];


      if (i==j) {
	if (modified) v = (s->values)[j] - dropped[j];


	v = v / pivot;

	(L->rowind)[next] = i;
	(L->values)[next] = v;
	next++;
	rowlist_add(i,j,v);
	break;
      }
    }

    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];
      
      if (i==j) continue;

      /* if (modified && i == j) v = (s->values)[j] - dropped[j]; */

      if (i==j || fabs(v) > droptol * norm || ip < Aj_nnz) {
	v = v / pivot;

	(L->rowind)[next] = i;
	(L->values)[next] = v;
	next++;
	rowlist_add(i,j,v);
      }
    }

    (L->colptr)[j+1] = next;
    {
      double Lj_nnz = (double) ((L->colptr)[j+1] - (L->colptr)[j]);
      flops += 2.0 * Lj_nnz * Lj_nnz;
    }

    /* rowlist_free(j); */
  }

  (L->colptr)[n] = next;
  
  rowlist_free();
  spa_free(s);
  free(dropped);

  taucs_printf("taucs_ccs_factor_llt: done; nnz(L) = %d, flops=%.1le\n",(L->colptr)[n],flops);

  return L;
}

/***************** FACTOR LLT PARTIAL ********************/

/* 
 * Partial LL^T factorization. Factors the first p columns
 * and then updates, but does not factor, the trailing submatrix.
 * Designed for Shur-complement preconditioning.
 * 
 */

/*
symccs_matrix* 
taucs_symccs_factor_llt_partial(symccs_matrix* A, 
                                int p) 
{return NULL;}
*/

taucs_ccs_matrix* 
taucs_ccs_factor_llt_partial(taucs_ccs_matrix* A, 
			     int p)
{
  int            i,j,k,l,n,ip,next,Lnnz;
  double Lkj,pivot,v;
  spa*           s;
  taucs_ccs_matrix* L;
  int Aj_nnz;
  double flops = 0.0;

  if (!(A->flags & TAUCS_SYMMETRIC)) {
    taucs_printf("taucs_ccs_factor_llt_partial: matrix must be symmetric\n");
    return NULL;
  }
  if (!(A->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_factor_llt_partial: lower part must be represented\n");
    return NULL;
  }

  n = A->n;

  taucs_printf("taucs_ccs_factor_llt_partial: starting n=%d p=%d\n",n,p);

  L = taucs_ccs_create(n,n,1000);
  L->flags = TAUCS_TRIANGULAR | TAUCS_LOWER;

  Lnnz = 1000;
  next = 0;

  s = spa_create(n);
  rowlist_create(n);

  for (j=0; j<p; j++) {
    spa_set(s,A,j);

    for (l = rowlist_getfirst(j); 
	 l != -1; 
	 l = rowlist_getnext(l)) {
      k   = rowlist_getcolind(l);
      Lkj = rowlist_getvalue(l);
      spa_scale_add(s,j,L,k,-Lkj); /*  L_*j -= L_kj * L_*k  */
    }

    /* we now add the j'th column of L to the symccs */
    
    if ( next+(s->length) > Lnnz ) {
      int*    rowind;
      double* values;
      int inc = max( 8192, s->length );
      
      Lnnz += inc;

      rowind = (int*)    realloc( L->rowind, Lnnz * sizeof(int) );
      values = (double*) realloc( L->values, Lnnz * sizeof(double) );
      /* check for errors */
      assert( rowind && values );
      L->rowind = rowind;
      L->values = values;
    }

    (L->colptr)[j] = next;

    Aj_nnz = (A->colptr)[j+1] - (A->colptr)[j]; 

    pivot = sqrt( (s->values)[j] );

    if (pivot == 0.0) {
      taucs_printf("taucs_ccs_factor_llt_partial: zero pivot in column %d\n",j);
    } else if (fabs(pivot) < 1e-12) {
      taucs_printf("taucs_ccs_factor_llt_partial: small pivot in column %d (%le)\n",j,pivot);
    }

    /* we want Lii to be first in the compressed column */
    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];

      if (i==j) {
	v = v / pivot;

	(L->rowind)[next] = i;
	(L->values)[next] = v;
	next++;
	rowlist_add(i,j,v);
	break;
      }
    }

    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];
      
      if (i==j) continue;

      v = v / pivot;

      (L->rowind)[next] = i;
      (L->values)[next] = v;
      next++;
      rowlist_add(i,j,v);
    }

    (L->colptr)[j+1] = next;

    {
      double Lj_nnz = (double) ((L->colptr)[j+1] - (L->colptr)[j]);
      flops += 2.0 * Lj_nnz * Lj_nnz;
    }
  }

  for (j=p; j<n; j++) {
    spa_set(s,A,j);

    /* we only apply updates from columns 0..p-1 */
    for (l = rowlist_getfirst(j); 
	 l != -1; 
	 l = rowlist_getnext(l)) {
      k   = rowlist_getcolind(l);
      Lkj = rowlist_getvalue(l);
      if (k >= p) continue; 
      spa_scale_add(s,j,L,k,-Lkj); /*  L_*j -= L_kj * L_*k  */
    }

    /* we now add the j'th column of L to the symccs */
    
    if ( next+(s->length) > Lnnz ) {
      int*    rowind;
      double* values;
      int inc = max( 8192, s->length );
      
      Lnnz += inc;

      rowind = (int*)    realloc( L->rowind, Lnnz * sizeof(int) );
      values = (double*) realloc( L->values, Lnnz * sizeof(double) );
      /* check for errors */
      assert( rowind && values );
      L->rowind = rowind;
      L->values = values;
    }

    (L->colptr)[j] = next;

    Aj_nnz = (A->colptr)[j+1] - (A->colptr)[j]; 

    /* we want Lii to be first in the compressed column */
    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];

      if (i==j) {
	(L->rowind)[next] = i;
	(L->values)[next] = v;
	next++;
	rowlist_add(i,j,v);
	break;
      }
    }

    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];
      
      if (i==j) continue;
      (L->rowind)[next] = i;
      (L->values)[next] = v;
      next++;
      rowlist_add(i,j,v);
    }

    (L->colptr)[j+1] = next;

    /* not sure the flop count is correct. */
    {
      double Lj_nnz = (double) ((L->colptr)[j+1] - (L->colptr)[j]);
      flops += 2.0 * Lj_nnz * Lj_nnz;
    }
  }

  (L->colptr)[n] = next;
  
  rowlist_free();
  spa_free(s);

  taucs_printf("taucs_ccs_factor_llt_partial: done; nnz(L) = %d, flops=%.1le\n",(L->colptr)[n],flops);

  return L;
}

/*********************************************************/
/* LDL^T factorization                                   */
/* This is a left-looking column-column code using       */
/* row lists.                                            */
/*********************************************************/

taucs_ccs_matrix* taucs_ccs_factor_ldlt(taucs_ccs_matrix* A)
{
  int            i,j,k,l,n,ip,next,Lnnz;
  double Lkj,pivot,v,norm,Dkk;
  spa*           s;
  taucs_ccs_matrix* L;
  int Aj_nnz;
  double flops = 0.0;

  n = A->n;

  taucs_printf("taucs_ccs_factor_ldlt: starting n=%d\n",n);

  L = taucs_ccs_create(n,n,1000);
  L->flags = TAUCS_TRIANGULAR | TAUCS_LOWER;

  Lnnz = 1000;
  next = 0;

  s = spa_create(n);
  rowlist_create(n);

  for (j=0; j<n; j++) {
    spa_set(s,A,j);

    for (l = rowlist_getfirst(j); 
	 l != -1; 
	 l = rowlist_getnext(l)) {
      k   = rowlist_getcolind(l);
      Lkj = rowlist_getvalue(l);
      Dkk = (L->values)[ (L->colptr)[k] ];
      spa_scale_add(s,j,L,k,-Lkj*Dkk); // L_*j -= L_kj * L_*k
    }

    /* we now add the j'th column of L to the taucs_ccs */
    
    if ( next+(s->length) > Lnnz ) {
      int*    rowind;
      double* values;
      int inc = max( 8192, s->length );
      
      Lnnz += inc;

      rowind = (int*)    realloc( L->rowind, Lnnz * sizeof(int) );
      values = (double*) realloc( L->values, Lnnz * sizeof(double) );
      /* check for errors */
      assert( rowind && values );
      L->rowind = rowind;
      L->values = values;
    }

    (L->colptr)[j] = next;

    Aj_nnz = (A->colptr)[j+1] - (A->colptr)[j]; 

    pivot = (s->values)[j];

    if (pivot == 0.0) {
      taucs_printf("ldlt: zero pivot in column %d\n",j);
      taucs_printf("ldlt: Ajj in spa = %lg Aj_nnz=%d\n",
		 (s->values)[j],Aj_nnz);
    }

    /* we want Lii to be first in the compressed column */
    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];

      if (i==j) {
	v = v / pivot;
	(L->rowind)[next] = i;
	(L->values)[next] = pivot; /* we put D on the diagonal */
	next++;
	rowlist_add(i,j,v);
	break;
      }
    }

    for (ip = 0; ip < s->length; ip++) {
      i = (s->indices)[ip];
      v = (s->values)[i];
      
      if (i==j) continue;

      v = v / pivot;
      
      (L->rowind)[next] = i;
      (L->values)[next] = v;
      next++;
      rowlist_add(i,j,v);
    }

    (L->colptr)[j+1] = next;
    {
      double Lj_nnz = (double) ((L->colptr)[j+1] - (L->colptr)[j]);
      flops += 2.0 * Lj_nnz * Lj_nnz;
    }

    //rowlist_free(j);
  }

  (L->colptr)[n] = next;
  
  rowlist_free();
  spa_free(s);

  taucs_printf("taucs_ccs_factor_ldlt: done; nnz(L) = %.2le, flops=%.2le\n",
	     (double) (L->colptr)[n],flops);

  return L;
}

/*********************************************************/
/*                                                       */
/*********************************************************/
